(*
Copyright 2009 Christian Sternagel, René Thiemann, Sarah Winkler, Harald Zankl

This file is part of IsaFoR/CeTA.

IsaFoR/CeTA is free software: you can redistribute it and/or modify it under the
terms of the GNU Lesser General Public License as published by the Free Software
Foundation, either version 3 of the License, or (at your option) any later
version.

IsaFoR/CeTA is distributed in the hope that it will be useful, but WITHOUT ANY
WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License along
with IsaFoR/CeTA. If not, see <http://www.gnu.org/licenses/>.
*)
theory DpFramework imports Trs Graph begin

text {* A \emph{dependency pair problem} (DP problem, for short) is a tuple consisting
of two rewrite systems *} 
types ('f,'v)dpp = "('f,'v)trel \<times> ('f,'v)trel"

text {* A \emph{graph-DP problem} is a tuple consisting of a graph (over DPs) and a rewrite system *}
types ('f,'v)gdpp = "('f,'v)rule graph \<times> ('f,'v)trel"

definition minimal_cond :: "('f,'v)trel \<Rightarrow> (nat \<Rightarrow> ('f,'v)term) \<Rightarrow> (nat \<Rightarrow> ('f,'v)term) \<Rightarrow> (nat \<Rightarrow> ('f,'v)sub) \<Rightarrow> bool"
where "minimal_cond R s t \<sigma> \<equiv> \<forall> i. (SN (rstep R) (t i \<cdot> \<sigma> i))"


fun min_graph_ichain ::
 "('f,'v)gdpp \<Rightarrow> (nat \<Rightarrow> ('f,'v)term) \<Rightarrow> (nat \<Rightarrow> ('f,'v)term) \<Rightarrow> (nat \<Rightarrow> ('f,'v)sub) \<Rightarrow> bool"
where "min_graph_ichain (P,R) s t \<sigma> = (
 (\<forall>i. (s i, t i) \<in> nodes P) \<and>
 (\<forall>i. ((s i,t i),(s(Suc i),t(Suc i))) \<in> edges P) \<and>
 (\<forall>i. ((t i)\<cdot>(\<sigma> i),(s(Suc i))\<cdot>(\<sigma> (Suc i))) \<in> (rstep R)^*) \<and>
 minimal_cond R s t \<sigma>
)"

fun ichain ::
 "('f,'v)dpp \<Rightarrow> (nat \<Rightarrow> ('f,'v)term) \<Rightarrow> (nat \<Rightarrow> ('f,'v)term) \<Rightarrow> (nat \<Rightarrow> ('f,'v)sub) \<Rightarrow> bool"
where "ichain (P,R) s t \<sigma> = (
 (\<forall>i. ((s i,t i) \<in> P)) \<and> 
 (\<forall>i. ((t i)\<cdot>(\<sigma> i),(s(Suc i))\<cdot>(\<sigma> (Suc i))) \<in> (rstep R)^*) 
)"

text {* A \emph{minimal infinite chain} is an infinite chain where additionally all @{text \<sigma>}-instances
of terms in the sequence~@{text t} are terminating w.r.t.~the system~@{text R}. *}
fun min_ichain ::
 "('f,'v)dpp \<Rightarrow> (nat \<Rightarrow> ('f,'v)term) \<Rightarrow> (nat \<Rightarrow> ('f,'v)term) \<Rightarrow> (nat \<Rightarrow> ('f,'v)sub) \<Rightarrow> bool"
where "min_ichain (P,R) s t \<sigma> = (ichain (P,R) s t \<sigma> \<and> minimal_cond R s t \<sigma>)"

lemma min_ichain_imp_ichain: assumes "min_ichain DPP s t \<sigma>" shows "ichain DPP s t \<sigma>"
using assms
by (cases DPP, auto)

fun shift :: "(nat \<Rightarrow> 'a) \<Rightarrow> nat \<Rightarrow> (nat \<Rightarrow> 'a)"
  where "shift f j = (\<lambda> i. f (i+j))"

lemma minimal_cond_mono : 
  assumes subset: "R \<subseteq> R'" and cond: "minimal_cond R' s t \<sigma>" 
  shows "minimal_cond R s t \<sigma>"
unfolding minimal_cond_def
proof 
  fix i
  from cond have SN: "SN (rstep R') (t i \<cdot> \<sigma> i)"
    unfolding minimal_cond_def by auto
  have "rstep R \<subseteq> rstep R'" by (rule rstep_mono, rule subset)
  with SN show "SN (rstep R) (t i \<cdot> \<sigma> i)" by (rule SN_imp_SN_subset)
qed


lemma min_ichainI[intro] : 
  assumes "R \<subseteq> R'" and "minimal_cond R' s t \<sigma>" and "ichain (P,R) (shift s i) (shift t i) (shift \<sigma> i)" 
  shows "min_ichain (P,R) (shift s i) (shift t i) (shift \<sigma> i)"
proof -
  from assms have "minimal_cond R s t \<sigma>" using minimal_cond_mono by auto
  hence "minimal_cond R (shift s i) (shift t i) (shift \<sigma> i)" 
    unfolding minimal_cond_def by auto
  with assms show ?thesis by auto
qed


fun initial_dpp :: "('f,'v)trel \<Rightarrow> ('f shp,'v)dpp"
where "initial_dpp R = (DP R, ID R)"

definition finite_dpp :: "('f,'v)dpp \<Rightarrow> bool" where "finite_dpp DPP = (\<not>(\<exists>s t \<sigma>. min_ichain DPP s t \<sigma>))"

definition finite_gdpp :: "('f,'v)gdpp \<Rightarrow> bool" where "finite_gdpp DPP = (\<not>(\<exists>s t \<sigma>. min_graph_ichain DPP s t \<sigma>))"

fun next4 :: "('f,'v)trel \<Rightarrow> ('f,'v)term \<Rightarrow>
 (('f shp,'v)term \<times> ('f shp,'v)term \<times> ('f,'v)term \<times> ('f shp,'v)sub)"
where "next4 R s = (SOME (l,r,u,\<sigma>).
 (l,r) \<in> DP R \<and> r\<cdot>\<sigma> = sharp u \<and> Tinf R u \<and> (sharp s,l\<cdot>\<sigma>) \<in> (rstep(ID R))^*)"

fun concrete_seq :: "('f,'v)trel \<Rightarrow> ('f,'v)term \<Rightarrow> nat \<Rightarrow>
 (('f,'v)term \<times> ('f shp,'v)term \<times> ('f shp,'v)term \<times> ('f,'v)term \<times> ('f shp,'v)sub)"
where "concrete_seq R s 0       = (s,next4 R s)"
    | "concrete_seq R r (Suc n) = (let (_,_,_,s,_) = concrete_seq R r n in (s,next4 R s))"

lemma someI_ex4: "\<exists>w x y z. P(w,x,y,z) \<Longrightarrow> P(SOME(w,x,y,z). P(w,x,y,z))" using some_eq_ex by force

theorem not_SN_imp_min_ichain:
 fixes R :: "('f,'v)trel"
 assumes "wf_trs R" and "\<not>(SN(rstep R))"
 shows "\<exists>s t \<sigma>. min_ichain (initial_dpp R) s t \<sigma>"
proof -
 from `\<not>(SN(rstep R))` obtain s' :: "('f,'v)term" where "\<not>(SN (rstep R) s')" unfolding SN_ars_def by auto
 then obtain s where "s' \<unrhd>  s" and "Tinf R s" using not_SN_imp_subt_Tinf by auto
 let ?I = "\<lambda>i. fst(concrete_seq R s i)"
 let ?S = "\<lambda>i. fst(snd(concrete_seq R s i))"
 let ?T = "\<lambda>i. fst(snd(snd(concrete_seq R s i)))"
 let ?\<sigma> = "\<lambda>i. snd(snd(snd(snd(concrete_seq R s i))))"
 let ?P = "\<lambda>s l r u \<sigma>. (l,r) \<in> DP R \<and> r\<cdot>\<sigma> = sharp u \<and> Tinf R u \<and> (sharp s,l\<cdot>\<sigma>) \<in> (rstep(ID R))^*"
 let ?P' =  "\<lambda>(l,r,u,\<sigma>). ?P s l r u \<sigma>"
 have main: "\<forall>i. ?P (?I i) (?S i) (?T i) (?I(Suc i)) (?\<sigma> i)"
 proof
  fix i show "?P (?I i) (?S i) (?T i) (?I(Suc i)) (?\<sigma> i)"
  proof (induct i)
   case 0 
   have "\<exists>l r u \<sigma>. ?P'(l,r,u,\<sigma>)" using Tinf_lemma_DP_explicit_sub[OF `wf_trs R` `Tinf R s`] by auto
   hence "?P'(next4 R s)" unfolding next4.simps using someI_ex4[where P = "?P'"] by best
   have ini: "?I 0 = s" unfolding concrete_seq.simps by simp
   have s: "?S 0 = fst(next4 R s)" unfolding concrete_seq.simps by simp
   have t: "?T 0 = fst(snd(next4 R s))" unfolding concrete_seq.simps by simp
   have sub: "?\<sigma> 0 = snd(snd(snd(next4 R s)))" unfolding concrete_seq.simps Let_def split_def by auto
   have nxt: "?I(Suc 0) = fst(snd(snd(next4 R s)))" unfolding concrete_seq.simps Let_def split_def by auto
   show ?case unfolding ini s t sub nxt using `?P'(next4 R s)` by auto
  next
   case (Suc i)
   let ?P'' = "\<lambda>(w,x,y,z). ?P (?I(Suc i)) w x y z"
   from Suc have "Tinf R (?I(Suc i))" by simp
   have "\<exists>l r u \<sigma>. ?P''(l,r,u,\<sigma>)" using Tinf_lemma_DP_explicit_sub[OF `wf_trs R` `Tinf R (?I(Suc i))`] by auto
   hence IH: "?P''(next4 R (?I(Suc i)))" unfolding next4.simps using someI_ex4[where P = "?P''"] by simp
   have s: "?S(Suc i) = fst(next4 R (?I(Suc i)))" unfolding concrete_seq.simps Let_def split_def by simp
   have t: "?T(Suc i) = fst(snd(next4 R (?I(Suc i))))" unfolding concrete_seq.simps Let_def split_def by simp
   have sub: "?\<sigma>(Suc i) = snd(snd(snd(next4 R (?I(Suc i)))))" unfolding concrete_seq.simps Let_def split_def by simp
   have nxt: "?I(Suc(Suc i)) = fst(snd(snd(next4 R (?I(Suc i)))))" unfolding concrete_seq.simps Let_def split_def by simp
   show ?case unfolding s t sub nxt using IH by auto
  qed
 qed
 from main have P: "\<forall>i. (?S i,?T i) \<in> DP R" by simp
 from main have eq: "\<And>i. sharp(?I(Suc i)) = ?T i \<cdot> ?\<sigma> i" by auto
 have R: "\<forall>i. (?T i \<cdot> ?\<sigma> i,?S(Suc i) \<cdot> ?\<sigma>(Suc i)) \<in> (rstep(ID R))^*"
 proof
  fix i show "(?T i \<cdot> ?\<sigma> i,?S(Suc i)\<cdot>?\<sigma>(Suc i)) \<in> (rstep(ID R))^*"
  proof (cases i)
   case 0
   from main have first: "?T 0 \<cdot> ?\<sigma> 0 = sharp(?I(Suc 0))" by best
   from main have "(sharp(?I(Suc 0)),?S(Suc 0)\<cdot>?\<sigma>(Suc 0)) \<in> (rstep(ID R))^*" by best
   thus ?thesis unfolding 0 unfolding first by simp
  next
   case (Suc i) show ?thesis unfolding eq[symmetric] using main by best
  qed
 qed
 from P and R have ichain: "ichain (initial_dpp R) ?S ?T ?\<sigma>" by auto
 have Tinf: "\<forall>i. Tinf R (?I i)"
 proof
  fix i show "Tinf R (?I i)" using main by (induct i) (auto simp: `Tinf R s`)
 qed
 have SN_t: "\<forall>i. SN (rstep(ID R)) (?T i \<cdot> ?\<sigma> i)"
 proof
  fix i
  from Tinf have "Tinf R (?I(Suc i))" by best
  from eq have eq': "sharp(?I(Suc i)) = ?T i \<cdot> ?\<sigma> i" by simp
  from Tinf_imp_SN_sharp[OF `wf_trs R` `Tinf R (?I(Suc i))`]
  show "SN (rstep(ID R)) (?T i \<cdot> ?\<sigma> i)" unfolding eq' by simp
 qed
 have SN_s: "\<forall>i. SN (rstep(ID R)) (?S i \<cdot> ?\<sigma> i)"
 proof
  fix i show "SN (rstep(ID R)) (?S i \<cdot> ?\<sigma> i)"
  proof (cases i)
   case 0
   have "?I 0 = s" by simp
   have SN_sharp: "SN (rstep(ID R)) (sharp s)" using Tinf_imp_SN_sharp[OF `wf_trs R` `Tinf R s`] by simp
   from main have "(sharp(?I 0),?S 0 \<cdot> ?\<sigma> 0) \<in> (rstep(ID R))^*" by best
   hence "(sharp s,?S 0 \<cdot> ?\<sigma> 0) \<in> (rstep(ID R))^*" by simp
   from rtrancl_step_preserves_SN[OF this SN_sharp] show ?thesis unfolding 0 by simp
  next
   case (Suc i)
   from SN_t have SN: "SN (rstep(ID R)) (?T i \<cdot> ?\<sigma> i)" by simp
   from R have "(?T i \<cdot> ?\<sigma> i,?S(Suc i) \<cdot> ?\<sigma>(Suc i)) \<in> (rstep(ID R))^*" by simp
   from rtrancl_step_preserves_SN[OF this SN] show ?thesis unfolding Suc by simp
  qed
 qed
 from ichain and SN_s and SN_t
 show ?thesis unfolding initial_dpp.simps min_ichain.simps minimal_cond_def by blast
qed
 
hide const concrete_seq next4

lemma min_graph_ichain_imp_min_ichain : assumes "min_graph_ichain (P,R) s t \<sigma>"
  shows "min_ichain (nodes P, R) s t \<sigma>"
using assms
by (simp add: nodes_def edges_def)

lemma gdpp_to_dpp : assumes fin: "finite_dpp (nodes P, R)" shows "finite_gdpp (P, R)"
  using assms min_graph_ichain_imp_min_ichain  
  by (unfold finite_dpp_def finite_gdpp_def, blast)
    
lemma dpp_lift_to_gdpp : fixes P :: "('f, 'v)rule graph"
  assumes 
  step: "\<And> s t \<sigma>. min_ichain (nodes P, R) s t \<sigma> \<Longrightarrow> \<exists> j. min_ichain (rP, R) (shift s j) (shift t j) (shift \<sigma> j)"
  and res: "finite_gdpp (node_restriction P rP, R)" (is "finite_gdpp (?rP, _)")
  shows "finite_gdpp (P, R)"
proof (unfold finite_gdpp_def, rule notI, clarify)
  fix s t \<sigma>
  assume chain: "min_graph_ichain (P, R) s t \<sigma>"
  hence "min_ichain (nodes P, R) s t \<sigma>" by (rule min_graph_ichain_imp_min_ichain)
  with step have "\<exists> j. min_ichain (rP, R) (shift s j) (shift t j) (shift \<sigma> j)" by simp
  from this obtain j where rPchain: "min_ichain (rP, R) (shift s j) (shift t j) (shift \<sigma> j)" ..
  with chain have "min_graph_ichain (?rP, R) (shift s j) (shift t j) (shift \<sigma> j)" 
  proof (simp add: node_restriction_def edges_def nodes_def, clarify)
    fix i
    assume "\<forall> i. (s (i+j), t (i + j)) \<in> rP"
    hence "(s (Suc i + j), t (Suc i + j)) \<in> rP" ..
    thus "(s (Suc (i+j)), t (Suc (i+j))) \<in> rP" by simp
  qed
  with res show False by (unfold finite_gdpp_def, blast)
qed

lemma finite_ini_imp_SN: assumes "wf_trs R" and "finite_dpp(initial_dpp R)" shows "SN(rstep R)"
proof (rule ccontr)
 assume "\<not>SN(rstep R)"
 from not_SN_imp_min_ichain[OF `wf_trs R` `\<not>(SN(rstep R))`]
 obtain s t \<sigma> where "min_ichain (initial_dpp R) s t \<sigma>" by best
 with `finite_dpp(initial_dpp R)` show False unfolding finite_dpp_def by best
qed

(* and a version which preprocesses the rules *)
fun computeRuleMap :: "('f,'v)trsL \<Rightarrow> ('f \<Rightarrow> ('f,'v)trsL) option"
where "computeRuleMap [] = Some (\<lambda> f. [])"
    | "computeRuleMap ((Fun f ts,r) # rules) = (case computeRuleMap rules of None \<Rightarrow> None | Some rm \<Rightarrow> 
            Some (rm(f := (Fun f ts,r) # (rm f))))"
    | "computeRuleMap ((Var _, _) # rules) = None"

lemma computeRuleMapSound : assumes "computeRuleMap R = Some rm"
  shows "(set (rm f)) = {(l,r) | l r. (l,r) \<in> set R \<and> root l = Some f}"
using assms 
proof (induct R arbitrary: rm f, force)
  case (Cons rule rules)
  show ?case
  proof (cases rule)
    case (Pair l r)
    show ?thesis
    proof (cases l)
      case (Var v)
      with Cons Pair show ?thesis by simp
    next
      case (Fun g ts)
      with Cons Pair show ?thesis by (cases "computeRuleMap rules", simp, cases "f = g", auto)
    qed
  qed
qed

lemma computeRuleMapSound2 : "(computeRuleMap R = None) = (\<exists> (l, r) \<in> set R. root l = None)"
using assms 
proof (induct R, simp)
  case (Cons rule rules)
  show ?case 
  proof (cases rule)
    case (Pair l r)
    with Cons show ?thesis by (cases l, simp, cases "computeRuleMap rules", auto)
  qed
qed


definition chain_identifying_proc :: "(('f,'v)trel \<Rightarrow> ('f,'v)trel \<Rightarrow> ('f,'v)trel \<Rightarrow> ('f,'v)trel \<Rightarrow> bool) \<Rightarrow> bool"
  where "chain_identifying_proc proc \<equiv> \<forall> P R rP rR. (proc P R rP rR \<longrightarrow> (rR \<subseteq> R \<and> (\<forall> s t \<sigma> . min_ichain (P,R) s t \<sigma> \<longrightarrow> (\<exists> i. ichain (rP, rR) (shift s i) (shift t i) (shift \<sigma> i)))))"

definition sound_proc 
  where "sound_proc proc \<equiv> \<forall> P R rP rR. (proc P R rP rR \<longrightarrow> (finite_dpp (rP, rR) \<longrightarrow> finite_dpp (P, R)))"

lemma ci_proc_to_sound_proc[simp] : 
  assumes ci_proc: "chain_identifying_proc proc"
  shows "sound_proc proc"
unfolding sound_proc_def
proof ((rule allI)+, (rule impI)+)
  fix P R rP rR
  assume cond: "proc P R rP rR" and finite: "finite_dpp (rP, rR)"
  show "finite_dpp (P,R)"
  proof (rule ccontr)
    assume "\<not> finite_dpp (P,R)"
    from this obtain s t \<sigma> where min: "min_ichain (P,R) s t \<sigma>" unfolding finite_dpp_def by auto
    from this cond ci_proc obtain i where subset: "rR \<subseteq> R" and ichain: "ichain (rP,rR) (shift s i) (shift t i) (shift \<sigma> i)" unfolding chain_identifying_proc_def by force
    have "min_ichain (rP, rR) (shift s i) (shift t i) (shift \<sigma> i)" 
      using min_ichainI[where R' = R and R = rR and P = rP] using ichain min subset by auto
    hence "\<not> finite_dpp (rP, rR)" unfolding finite_dpp_def by blast
    with finite show False by auto
  qed
qed

definition chain_identifying_proc_impl :: "(('f,'v)trsL \<Rightarrow> ('f,'v)trsL \<Rightarrow> ('f \<Rightarrow> ('f,'v)trsL)option \<Rightarrow> ('f,'v)trsL \<Rightarrow> ('f,'v)trsL \<Rightarrow> bool) \<Rightarrow> bool"
where "chain_identifying_proc_impl proc \<equiv> \<forall> P R rP rR. (proc P R (computeRuleMap R) rP rR \<longrightarrow> (set rR \<subseteq> set R \<and> (\<forall> s t \<sigma> . min_ichain (set P, set R) s t \<sigma> \<longrightarrow> (\<exists> i. ichain (set rP, set rR) (shift s i) (shift t i) (shift \<sigma> i)))))"

definition sound_proc_impl :: "(('f,'v)trsL \<Rightarrow> ('f,'v)trsL \<Rightarrow> ('f \<Rightarrow> ('f,'v)trsL)option \<Rightarrow> ('f,'v)trsL \<Rightarrow> ('f,'v)trsL \<Rightarrow> bool) \<Rightarrow> bool"
where "sound_proc_impl proc \<equiv> \<forall> P R rP rR. (proc P R (computeRuleMap R) rP rR \<longrightarrow> (finite_dpp (set rP, set rR) \<longrightarrow> finite_dpp (set P, set R)))"

definition chain_identifying_onlyDP_proc_impl :: "(('f,'v)trsL \<Rightarrow> ('f,'v)trsL \<Rightarrow> ('f \<Rightarrow> ('f,'v)trsL)option \<Rightarrow> ('f,'v)trsL \<Rightarrow> bool) \<Rightarrow> bool"
where "chain_identifying_onlyDP_proc_impl proc \<equiv> \<forall> P R rP. (proc P R (computeRuleMap R) rP \<longrightarrow> ((\<forall> s t \<sigma> . min_ichain (set P, set R) s t \<sigma> \<longrightarrow> (\<exists> i. ichain (set rP, set R) (shift s i) (shift t i) (shift \<sigma> i)))))"

definition sound_onlyDP_proc_impl :: "(('f,'v)trsL \<Rightarrow> ('f,'v)trsL \<Rightarrow> ('f \<Rightarrow> ('f,'v)trsL)option \<Rightarrow> ('f,'v)trsL \<Rightarrow> bool) \<Rightarrow> bool"
where "sound_onlyDP_proc_impl proc \<equiv> \<forall> P R rP. (proc P R (computeRuleMap R) rP \<longrightarrow> (finite_dpp (set rP, set R) \<longrightarrow> finite_dpp (set P, set R)))"

lemma ci_onlyDP_imp_sound_onlyDP[simp] : 
  assumes ci_proc: "chain_identifying_onlyDP_proc_impl proc"
  shows "sound_onlyDP_proc_impl proc" 
unfolding sound_onlyDP_proc_impl_def
proof ((rule allI)+, (rule impI)+)
  fix P R rP 
  assume cond: "proc P R (computeRuleMap R) rP" and finite: "finite_dpp (set rP, set R)"
  show "finite_dpp (set P, set R)"
  proof (rule ccontr)
    assume "\<not> finite_dpp (set P, set R)"
    from this obtain s t \<sigma> where min: "min_ichain (set P, set R) s t \<sigma>" unfolding finite_dpp_def by auto
    from this cond ci_proc obtain i where ichain: "ichain (set rP,set R) (shift s i) (shift t i) (shift \<sigma> i)" unfolding chain_identifying_onlyDP_proc_impl_def by blast
    have "min_ichain (set rP, set R) (shift s i) (shift t i) (shift \<sigma> i)" 
      using min_ichainI[where R' = "set R" and R = "set R" and P = "set rP"] using ichain min by auto
    hence "\<not> finite_dpp (set rP, set R)" unfolding finite_dpp_def by blast
    with finite show False by auto
  qed
qed

lemma apply_ci_onlyDP_proc : assumes ci: "chain_identifying_onlyDP_proc_impl proc"
  and fin: "finite_dpp (set rP, set R)"
  and proc: "proc P R (computeRuleMap R) rP"
  shows "finite_dpp (set P, set R)"
proof -
  from ci have "sound_onlyDP_proc_impl proc" by simp
  thus ?thesis unfolding sound_onlyDP_proc_impl_def using fin proc by blast
qed

lemma ci_imp_sound[simp] : 
  assumes ci_proc: "chain_identifying_proc_impl proc"
  shows "sound_proc_impl proc" 
unfolding sound_proc_impl_def
proof ((rule allI)+, (rule impI)+)
  fix P R rP rR
  assume cond: "proc P R (computeRuleMap R) rP rR" and finite: "finite_dpp (set rP, set rR)"
  show "finite_dpp (set P, set R)"
  proof (rule ccontr)
    assume "\<not> finite_dpp (set P, set R)"
    from this obtain s t \<sigma> where min: "min_ichain (set P, set R) s t \<sigma>" unfolding finite_dpp_def by auto
    from this cond ci_proc obtain i where ichain: "set rR \<subseteq> set R \<and> ichain (set rP,set rR) (shift s i) (shift t i) (shift \<sigma> i)" unfolding chain_identifying_proc_impl_def by blast
    have "min_ichain (set rP, set rR) (shift s i) (shift t i) (shift \<sigma> i)" 
      using min_ichainI[where R' = "set R" and R = "set rR" and P = "set rP"] using ichain min by auto
    hence "\<not> finite_dpp (set rP, set rR)" unfolding finite_dpp_def by blast
    with finite show False by auto
  qed
qed

lemma apply_ci_proc : assumes ci: "chain_identifying_proc_impl proc"
  and fin: "finite_dpp (set rP, set rR)"
  and proc: "proc P R (computeRuleMap R) rP rR"
  shows "finite_dpp (set P, set R)"
proof -
  from ci have "sound_proc_impl proc" by simp
  thus ?thesis unfolding sound_proc_impl_def using fin proc by blast
qed


end
